

#include <assert.h>
#include <cublas.h>
#include <iostream>

#include "cutil.h"

using namespace std;

#define real double

void cublas_dgemm_test(int I, int J, int K) 
{
  // compute C=A'*B using cublasDgemm, and compare to the expected
  // result. A is KxI, B is KxJ, so C will be IxJ.  Matrices are
  // stored in column-major order.
  real *h_A, *h_B, *h_C;
  real *d_A, *d_B, *d_C;
  int A_size = K*I*sizeof(real);
  int B_size = K*J*sizeof(real);
  int C_size = I*J*sizeof(real);

  cout << "(" << I << ", " << J << ", " << K << "): "; 
  
  h_A = (double*)malloc(A_size);
  h_B = (double*)malloc(B_size);
  h_C = (double*)malloc(C_size);
  assert(h_C!=NULL);
  CUDA_SAFE_CALL( cudaMalloc((void**)&d_A, A_size) );
  CUDA_SAFE_CALL( cudaMalloc((void**)&d_B, B_size) );
  CUDA_SAFE_CALL( cudaMalloc((void**)&d_C, C_size) );
  for (int i=0; i<I; i++) {
    for (int k=0; k<K; k++) {
      h_A[K*i+k]=0.2;
    }
  }
  for (int j=0; j<J; j++) {
    for (int k=0; k<K; k++) {
      h_B[K*j+k]=1.;
    }
  }
  CUDA_SAFE_CALL( cudaMemcpy(d_A, h_A, A_size, cudaMemcpyHostToDevice) ); 
  CUDA_SAFE_CALL( cudaMemcpy(d_B, h_B, B_size, cudaMemcpyHostToDevice) );
  cublasDgemm('t','n', I, J, K, 1, d_A, K, d_B, K, 0, d_C, I);

  int d,e;
  d=cublasGetError();
  e=cudaGetLastError();
  if (d!=0 || e!=0) {
    cout << "CUDA/CUBLAS error in cublasDgemm" << endl;
    exit(-1);
  }
  CUDA_SAFE_CALL( cudaMemcpy(h_C, d_C, C_size, cudaMemcpyDeviceToHost) );

  // every entry of C should equal K*1.*0.2
  double expected_value = K*0.2;
  for (int j=0; j<J; j++) {
    for (int i=0; i<I; i++) {
      double dgemm_value = h_C[I*j+i];
      if (fabs(dgemm_value-expected_value)>1e-6) {
      	cout << "Fail!" << endl;      
	cout << "Difference at C position (" << i << ", " << j << ")" << endl;
	cout << "Expected: " << expected_value << endl;
	cout << "Found: "    << dgemm_value    << endl;
	return;
      }
    } 
  }
  cout << "OK." << endl;
}


int main(int argc, char** argv) {
  int err = cublasInit();
  assert(err==0);
  
  cublas_dgemm_test(10,10,8);
  cublas_dgemm_test(100,100,8);
  cublas_dgemm_test(300,100,8);
  cublas_dgemm_test(300,1000,8);
  cublas_dgemm_test(300,10000,8);

  cublasShutdown();
} 
